/**
 * @file
 *
 * Task queue implementation.  This is inspired by the ``bounded
 * partial queue'' in Maurice Herlihy's book (Chapter 10.2).
 *
 * @author Jens Teubner <jens.teubner@cs.tu-dortmund.de>
 *
 * $Id: task_queue.c 3101 2017-07-03 12:06:41Z jens.teubner $
 */

#include "config.h"

#include <assert.h>

#include <stdbool.h>

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include "task_queue.h"


task_queue_t *
task_queue_init (const unsigned int capacity)
{
    task_queue_t *ret;
    int           status;

    ret = malloc (sizeof (*ret));
    if (!ret)
    {
        fprintf (stderr, "Error during memory allocation.\n");
        exit (EXIT_FAILURE);
    }

    status = pthread_mutex_init (&ret->enqLock, NULL);
    assert (!status);

    status = pthread_mutex_init (&ret->deqLock, NULL);
    assert (!status);

    status = pthread_cond_init (&ret->notEmptyCondition, NULL);
    assert (!status);

    status = pthread_cond_init (&ret->notFullCondition, NULL);
    assert (!status);

    ret->head = ret->tail = 0;

    ret->capacity = capacity;

    ret->terminateOnEmpty = false;

    ret->entries = malloc (capacity * sizeof (*ret->entries));
    if (!ret->entries)
    {
        fprintf (stderr, "Error during memory allocation.\n");
        exit (EXIT_FAILURE);
    }

    return ret;
}

void
task_queue_enq (task_queue_t *queue, void *x)
{
    bool mustWakeDequeuers = false;
    int  status;

    status = pthread_mutex_lock (&queue->enqLock);
    assert (!status);

    /* queue may be full; wait until there's space available */
    while (((queue->tail + 1) % queue->capacity) == queue->head)
    {
        status = pthread_cond_wait (&queue->notFullCondition, &queue->enqLock);
        assert (!status);
    }

    /* put user item into queue */
    queue->entries[queue->tail] = x;

    mustWakeDequeuers = (queue->tail == queue->head);

    queue->tail = (queue->tail + 1) % queue->capacity;

    status = pthread_mutex_unlock (&queue->enqLock);
    assert (!status);

    if (mustWakeDequeuers)
    {
        status = pthread_mutex_lock (&queue->deqLock);
        assert (!status);
        status = pthread_cond_signal (&queue->notEmptyCondition);
        assert (!status);
        status = pthread_mutex_unlock (&queue->deqLock);
        assert (!status);
    }
}

void *
task_queue_deq (task_queue_t *queue)
{
    void *result;
    bool  mustWakeEnqueuers = true;
    int   status;

    status = pthread_mutex_lock (&queue->deqLock);
    assert (!status);

    while (queue->tail == queue->head)
    {
        status = pthread_cond_wait (&queue->notEmptyCondition, &queue->deqLock);
        assert (!status);
    }

    result = (void *) queue->entries[queue->head];

    mustWakeEnqueuers = ((queue->tail + 1) % queue->capacity == queue->head);

    queue->head = (queue->head + 1) % queue->capacity;

    status = pthread_mutex_unlock (&queue->deqLock);

    if (mustWakeEnqueuers)
    {
        status = pthread_mutex_lock (&queue->enqLock);
        assert (!status);
        status = pthread_cond_signal (&queue->notFullCondition);
        assert (!status);
        status = pthread_mutex_unlock (&queue->enqLock);
        assert (!status);
    }

    return result;
}

/**
 * Dequeue, but only if there's something in the queue.  Otherwise
 * return @c NULL.
 */
void *
task_queue_trydeq (task_queue_t *queue)
{
    void *result;
    bool  mustWakeEnqueuers = true;
    int   status;

    status = pthread_mutex_lock (&queue->deqLock);
    assert (!status);

    if (queue->tail == queue->head)
    {
        pthread_mutex_unlock (&queue->deqLock);
        return NULL;
    }

    result = (void *) queue->entries[queue->head];

    mustWakeEnqueuers = ((queue->tail + 1) % queue->capacity == queue->head);

    queue->head = (queue->head + 1) % queue->capacity;

    status = pthread_mutex_unlock (&queue->deqLock);

    if (mustWakeEnqueuers)
    {
        status = pthread_mutex_lock (&queue->enqLock);
        assert (!status);
        status = pthread_cond_signal (&queue->notFullCondition);
        assert (!status);
        status = pthread_mutex_unlock (&queue->enqLock);
        assert (!status);
    }

    return result;
}
